from __future__ import annotations

import copy
import math
from typing import TYPE_CHECKING, Any, Literal, overload

from loguru import logger

from ..assets.data import TextMap
from ..assets.zzz.manager import ZZZ_ASSETS
from ..calc.zzz import LayerGenerator, PropState
from ..constants.common import DEFAULT_TIMEOUT, ZZZ_API_URL
from ..enums import zzz as enums
from ..errors import AssetKeyError, WrongUIDFormatError
from ..models import zzz as models
from ..models.zzz.build import Build
from .base import BaseClient

if TYPE_CHECKING:
    from ..models.enka.owner import Owner, OwnerInput
    from .cache import BaseTTLCache

__all__ = ("ZZZClient",)


class ZZZClient(BaseClient):
    """The main client to interact with the Enka Network Zenless Zone Zero API.

    Args:
        lang (Language | str): The language to use for the client, defaults to Language.ENGLISH.
        headers (dict[str, Any] | None): The headers to use for the client, defaults to None.
        cache (BaseTTLCache | None): The cache to use for the client, defaults to None.
        timeout (int): The timeout for the client, defaults to DEFAULT_TIMEOUT.
    """

    def __init__(
        self,
        lang: enums.Language | str = enums.Language.ENGLISH,
        *,
        headers: dict[str, Any] | None = None,
        cache: BaseTTLCache | None = None,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> None:
        super().__init__(headers=headers, cache=cache, timeout=timeout)

        self._lang = self._convert_lang(lang)
        self._assets = ZZZ_ASSETS

    def _convert_lang(self, lang: enums.Language | str) -> enums.Language:
        if isinstance(lang, str):
            try:
                lang = enums.Language(lang)
            except ValueError as e:
                available_langs = ", ".join(lang.value for lang in enums.Language)
                msg = f"Invalid language: {lang}, must be one of {available_langs}"
                raise ValueError(msg) from e

        return lang

    @property
    def lang(self) -> enums.Language:
        return self._lang

    @lang.setter
    def lang(self, lang: enums.Language | str) -> None:
        self._lang = self._convert_lang(lang)
        self._text_map = TextMap(self._lang, self._assets.text_map)

    async def __aenter__(self) -> ZZZClient:
        await self.start()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        return await super().__aexit__(exc_type, exc_val, exc_tb)

    async def start(self) -> None:
        await super().start()
        await self._assets.load(self.session)
        self._text_map = TextMap(self.lang, self._assets.text_map)

    async def update_assets(self) -> None:
        """Update game assets."""
        logger.info("Updating ZZZ assets")
        await self._assets.update(self.session)
        self._text_map = TextMap(self.lang, self._assets.text_map)
        logger.info("ZZZ assets updated")

    def _calc_disc_stats(self, disc: models.DriveDisc) -> None:
        # See https://api.enka.network/#/docs/zzz/api?id=drive-disc for the formula
        disc_level_data = next(
            (
                d
                for d in self._assets.equipment_level["Items"]
                if d["Level"] == disc.level and d["Rarity"] == disc.rarity_num
            ),
            None,
        )
        if disc_level_data is None:
            logger.warning(
                f"Drive disc level data not found for level={disc.level} and rarity={disc.rarity_num}"
            )
            return

        main_stat_val = disc.main_stat.value * (1 + disc_level_data["EnhanceRate"] / 10000)
        disc.main_stat.value = math.floor(main_stat_val)

        main_stat_data = self._assets.property[str(disc.main_stat.type.value)]
        disc.main_stat.name = self._text_map[main_stat_data["Name"]]
        disc.main_stat.format = main_stat_data["Format"]

        for sub in disc.sub_stats:
            sub.value *= sub.roll_times

            sub_data = self._assets.property[str(sub.type.value)]
            sub.name = self._text_map[sub_data["Name"]]
            sub.format = sub_data["Format"]

    def _calc_engine_stats(self, engine: models.WEngine) -> None:
        # See https://api.enka.network/#/docs/zzz/api?id=w-engine for the formula

        level_data = next(
            (
                d
                for d in self._assets.weapon_level["Items"]
                if d["Level"] == engine.level and d["Rarity"] == engine.rarity_num
            ),
            None,
        )
        if level_data is None:
            logger.warning(
                f"W-Engine level data not found for level={engine.level} and rarity={engine.rarity_num}"
            )
            return

        enhance_rate = level_data["EnhanceRate"]

        star_data = next(
            (
                d
                for d in self._assets.weapon_star["Items"]
                if d["Rarity"] == engine.rarity_num and d["BreakLevel"] == engine.modification
            ),
            None,
        )
        if star_data is None:
            logger.warning(f"W-Engine star data not found for rarity={engine.rarity_num}")
            return

        star_rate = star_data["StarRate"]
        rand_rate = star_data["RandRate"]

        main_stat_val = engine.main_stat.value * (1 + enhance_rate / 10000 + star_rate / 10000)
        engine.main_stat.value = math.floor(main_stat_val)

        sub_stat_val = engine.sub_stat.value * (1 + rand_rate / 10000)
        engine.sub_stat.value = math.floor(sub_stat_val)

    def _post_process_engine(self, engine: models.WEngine) -> None:
        engine_data = self._assets.weapons[str(engine.id)]
        engine.name = self._text_map[engine_data["ItemName"]]
        engine.rarity_num = engine_data["Rarity"]
        engine.specialty = enums.ProfessionType(engine_data["ProfessionType"])
        engine.icon = f"https://enka.network{engine_data['ImagePath']}"

        main_stat = engine_data["MainStat"]
        main_stat_data = self._assets.property[str(main_stat["PropertyId"])]
        engine.main_stat = models.Stat(
            type=enums.StatType(main_stat["PropertyId"]),
            value=main_stat["PropertyValue"],
            name=self._text_map[main_stat_data["Name"]],
            format=main_stat_data["Format"],
        )

        sub_stat = engine_data["SecondaryStat"]
        sub_stat_data = self._assets.property[str(sub_stat["PropertyId"])]
        engine.sub_stat = models.Stat(
            type=enums.StatType(sub_stat["PropertyId"]),
            value=sub_stat["PropertyValue"],
            name=self._text_map[sub_stat_data["Name"]],
            format=sub_stat_data["Format"],
        )

        self._calc_engine_stats(engine)

    def _post_process_disc(self, disc: models.DriveDisc) -> None:
        disc_data = self._assets.equipments["Items"][str(disc.id)]
        disc.rarity_num = disc_data["Rarity"]
        disc.set_id = disc_data["SuitId"]
        self._calc_disc_stats(disc)

    def _post_process_agent(self, agent: models.Agent) -> None:
        data = self._assets.avatars[str(agent.id)]
        agent.name = self._text_map[data["Name"]]
        agent.rarity_num = data["Rarity"]
        agent.elements = [enums.Element(e) for e in data["ElementTypes"]]
        agent.icon = models.AgentIcon(
            filename=data["Image"].removeprefix("/ui/zzz/").removesuffix(".png")
        )
        agent.sig_engine_id = data["WeaponId"]
        agent.color = models.AgentColor(**data["Colors"])
        agent.highlight_stats = [enums.StatType(stat) for stat in data["HighlightProps"]]
        agent.specialty = enums.ProfessionType(data["ProfessionType"])

        for disc in agent.discs:
            self._post_process_disc(disc)

        if agent.w_engine is not None:
            self._post_process_engine(agent.w_engine)

        # Credit to Enka Network for the stats calculation logic
        generator = LayerGenerator(self._assets)
        prop_state = PropState()

        prop_state.add(generator.character(agent))
        prop_state.add(generator.core(agent))
        if agent.w_engine is not None:
            prop_state.add(generator.weapon(agent.w_engine))

        prop_state.add(generator.discs(agent.discs))
        prop_state.add(generator.disc_sets(agent.discs))
        prop_state.add(generator.corrections(agent, prop_state))

        props = prop_state.sum()

        final_stats: dict[enums.AgentStatType, float] = {
            enums.AgentStatType.MAX_HP: props.max_hp,
            enums.AgentStatType.ATK: props.atk,
            enums.AgentStatType.DEF: props.defense,
            enums.AgentStatType.IMPACT: props.break_stun,
            enums.AgentStatType.CRIT_RATE: props.crit,
            enums.AgentStatType.CRIT_DMG: props.crit_dmg,
            enums.AgentStatType.PEN_RATIO: props.pen_ratio,
            enums.AgentStatType.PEN: props.pen_delta,
            enums.AgentStatType.ENERGY_REGEN: props.sp_recover,
            enums.AgentStatType.SHEER_FORCE: props.skip_def_atk,
            enums.AgentStatType.AAA: props.rp_recover,
            enums.AgentStatType.ANOMALY_PROFICIENCY: props.element_mystery,
            enums.AgentStatType.ANOMALY_MASTERY: props.element_abnormal_power,
            enums.AgentStatType.PHYSICAL_DMG_BONUS: props.added_damage_ratio_physics,
            enums.AgentStatType.FIRE_DMG_BONUS: props.added_damage_ratio_fire,
            enums.AgentStatType.ICE_DMG_BONUS: props.added_damage_ratio_ice,
            enums.AgentStatType.ELECTRIC_DMG_BONUS: props.added_damage_ratio_elec,
            enums.AgentStatType.ETHER_DMG_BONUS: props.added_damage_ratio_ether,
            enums.AgentStatType.SHEER_DMG_BONUS: props.skip_def_damage_ratio,
        }

        agent.stats = {
            stat_type: models.AgentStat(
                type=stat_type,
                value=math.floor(value),
                name=self._text_map[self._assets.property[str(stat_type.value)]["Name"]],
                format=self._assets.property[str(stat_type.value)]["Format"],
            )
            for stat_type, value in final_stats.items()
        }

    def _post_process_title(self, title: models.Title) -> None:
        title_data = self._assets.titles[str(title.id)]
        title.text = self._text_map[title_data["TitleText"]]
        title.color1 = f"#{title_data['ColorA']}"
        title.color2 = f"#{title_data['ColorB']}"

    def _post_process_player(self, player: models.Player) -> None:
        self._post_process_title(player.title)

        # Avatar
        pfp_data = self._assets.pfps.get(str(player.id))
        if pfp_data:
            player.avatar = f"https://enka.network{pfp_data['Icon']}"

        # Namecard
        namecard_id = player.namecard_id
        try:
            icon_path = self._assets.namecards[str(namecard_id)]["Icon"]
            icon_url = f"https://enka.network{icon_path}"
        except AssetKeyError:
            icon_url = ""
        else:
            player.namecard = models.Namecard(id=namecard_id, icon=icon_url)

    def _post_process_showcase(self, showcase: models.ShowcaseResponse) -> None:
        self._post_process_player(showcase.player)
        for agent in showcase.agents:
            self._post_process_agent(agent)

    @overload
    async def fetch_showcase(
        self, uid: str | int, *, raw: Literal[False] = False
    ) -> models.ShowcaseResponse: ...
    @overload
    async def fetch_showcase(
        self, uid: str | int, *, raw: Literal[True] = True
    ) -> dict[str, Any]: ...
    async def fetch_showcase(
        self, uid: str | int, *, raw: bool = False
    ) -> models.ShowcaseResponse | dict[str, Any]:
        """Fetch the player showcase of the given UID.

        Args:
            uid: The UID of the user.
            raw: Whether to return the raw data, defaults to False.

        Returns:
            The parsed or raw showcase data.
        """
        if not str(uid).isdigit():
            raise WrongUIDFormatError

        url = ZZZ_API_URL.format(uid)
        data = await self._request(url)
        if raw:
            return data

        data = copy.deepcopy(data)
        showcase = models.ShowcaseResponse(**data)
        self._post_process_showcase(showcase)
        return showcase

    def parse_showcase(self, data: dict[str, Any]) -> models.ShowcaseResponse:
        """Parse the given showcase data.

        Args:
            data: The showcase data.

        Returns:
            The parsed showcase response.
        """
        data = copy.deepcopy(data)
        showcase = models.ShowcaseResponse(**data)
        self._post_process_showcase(showcase)
        return showcase

    async def fetch_builds(self, owner: Owner | OwnerInput) -> dict[str, list[Build]]:
        """Fetch the character builds of the given owner.

        Args:
            owner: The owner of the builds.

        Returns:
            Character ID to list of builds mapping.
        """
        data = await self._request_profile(owner)
        result: dict[str, list[Build]] = {}

        for key, builds in data.items():
            result[key] = []
            for build in builds:
                build_ = Build(**build)
                self._post_process_agent(build_.character)
                result[key].append(build_)

        return result



from __future__ import annotations

import math

from enka.constants.common import CN_UID_PREFIXES


def round_down(number: int | float, decimal_places: int) -> int | float:
    """Round down a number to a specified number of decimal places.

    Args:
        number: The number to round down.
        decimal_places: The number of decimal places to round down to.

    Returns:
        int | float: The rounded down number.
    """
    factor = 10**decimal_places
    return math.floor(number * factor) / factor


def is_hsr_cn_uid(uid: str) -> bool:
    """Check if the given HSR UID belongs to a Chinese server.

    Args:
        uid: The UID to check.

    Returns:
        bool: True if the UID is a CN UID, False otherwise.
    """
    # The length check is to avoid cases where the UID length
    # got extended like in Genshin (18... is Asia server)
    return uid.startswith(CN_UID_PREFIXES) and len(uid) == 9



from __future__ import annotations

from typing import Final

CN_UID_PREFIXES: Final[tuple[str, ...]] = ("1", "2", "5")
DEFAULT_TIMEOUT: Final[int] = 5  # Seconds before API request times out

HSR_API_URL: Final[str] = "https://enka.network/api/hsr/uid/{}"
HSR_BACKUP_API_URL: Final[str] = "https://api.asterity.net/hsr/uid/{}?parse=enka"
GI_API_URL: Final[str] = "https://enka.network/api/uid/{}"
ZZZ_API_URL: Final[str] = "https://enka.network/api/zzz/uid/{}"
PROFILE_API_URL: Final[str] = "https://enka.network/api/profile/{}/hoyos/{}/builds/"



from __future__ import annotations

import math
from collections import defaultdict
from typing import TYPE_CHECKING, Final

from ..enums.zzz import ProfessionType

if TYPE_CHECKING:
    from enka.assets.zzz.manager import ZZZAssetManager
    from enka.models.zzz.character import Agent, DriveDisc, WEngine

DEFAULT_PROPS: Final[dict[str, float]] = {
    "HpMax_Base": 0,
    "HpMax_Ratio": 0,
    "HpMax_Delta": 0,
    "Atk_Base": 0,
    "Atk_Ratio": 0,
    "Atk_Delta": 0,
    "BreakStun_Base": 0,
    "BreakStun_Ratio": 0,
    "SkipDefAtk_Base": 0,
    "SkipDefAtk_Delta": 0,
    "Def_Base": 0,
    "Def_Ratio": 0,
    "Def_Delta": 0,
    "Crit_Base": 0,
    "Crit_Delta": 0,
    "CritDmg_Base": 0,
    "CritDmg_Delta": 0,
    "PenRatio_Base": 0,
    "PenRatio_Delta": 0,
    "PenDelta_Base": 0,
    "PenDelta_Delta": 0,
    "SpRecover_Base": 0,
    "SpRecover_Ratio": 0,
    "SpRecover_Delta": 0,
    "ElementMystery_Base": 0,
    "ElementMystery_Delta": 0,
    "ElementAbnormalPower_Base": 0,
    "ElementAbnormalPower_Ratio": 0,
    "ElementAbnormalPower_Delta": 0,
    "AddedDamageRatio_Physics_Base": 0,
    "AddedDamageRatio_Physics_Delta": 0,
    "AddedDamageRatio_Fire_Base": 0,
    "AddedDamageRatio_Fire_Delta": 0,
    "AddedDamageRatio_Ice_Base": 0,
    "AddedDamageRatio_Ice_Delta": 0,
    "AddedDamageRatio_Elec_Base": 0,
    "AddedDamageRatio_Elec_Delta": 0,
    "AddedDamageRatio_Ether_Base": 0,
    "AddedDamageRatio_Ether_Delta": 0,
    "RpRecover_Base": 0,
    "RpRecover_Ratio": 0,
    "RpRecover_Delta": 0,
    "SkipDefDamageRatio_Base": 0,
    "SkipDefDamageRatio_Delta": 0,
}

PROP_ID_TO_NAME: Final[dict[int, str]] = {
    11101: "HpMax_Base",
    11102: "HpMax_Ratio",
    11103: "HpMax_Delta",
    12101: "Atk_Base",
    12102: "Atk_Ratio",
    12103: "Atk_Delta",
    12201: "BreakStun_Base",
    12202: "BreakStun_Ratio",
    12301: "SkipDefAtk_Base",
    12303: "SkipDefAtk_Delta",
    13101: "Def_Base",
    13102: "Def_Ratio",
    13103: "Def_Delta",
    20101: "Crit_Base",
    20103: "Crit_Delta",
    21101: "CritDmg_Base",
    21103: "CritDmg_Delta",
    23101: "PenRatio_Base",
    23103: "PenRatio_Delta",
    23201: "PenDelta_Base",
    23203: "PenDelta_Delta",
    30501: "SpRecover_Base",
    30502: "SpRecover_Ratio",
    30503: "SpRecover_Delta",
    31201: "ElementMystery_Base",
    31203: "ElementMystery_Delta",
    31401: "ElementAbnormalPower_Base",
    31402: "ElementAbnormalPower_Ratio",
    31403: "ElementAbnormalPower_Delta",
    31501: "AddedDamageRatio_Physics_Base",
    31503: "AddedDamageRatio_Physics_Delta",
    31601: "AddedDamageRatio_Fire_Base",
    31603: "AddedDamageRatio_Fire_Delta",
    31701: "AddedDamageRatio_Ice_Base",
    31703: "AddedDamageRatio_Ice_Delta",
    31801: "AddedDamageRatio_Elec_Base",
    31803: "AddedDamageRatio_Elec_Delta",
    31901: "AddedDamageRatio_Ether_Base",
    31903: "AddedDamageRatio_Ether_Delta",
    32001: "RpRecover_Base",
    32002: "RpRecover_Ratio",
    32003: "RpRecover_Delta",
    32201: "SkipDefDamageRatio_Base",
    32203: "SkipDefDamageRatio_Delta",
}


class LayerGenerator:
    def __init__(self, assets: ZZZAssetManager) -> None:
        self._assets = assets

    def character(self, a: Agent) -> PropLayer:
        layer = PropLayer()
        excel = self._assets.avatars.data[str(a.id)]

        for prop_id, base in excel["BaseProps"].items():
            growth = (excel["GrowthProps"].get(prop_id, 0) / 10000) * (a.level - 1)
            try:
                promotion = excel["PromotionProps"][a.promotion - 1][prop_id]
            except (IndexError, KeyError):
                promotion = 0

            computed = base + growth + promotion
            layer.add(PROP_ID_TO_NAME[int(prop_id)], computed)

        return layer

    def core(self, a: Agent) -> PropLayer:
        layer = PropLayer()
        excel = self._assets.avatars.data[str(a.id)]
        props = excel["CoreEnhancementProps"][a.core_skill_level_num]

        for prop_id, core in props.items():
            layer.add(PROP_ID_TO_NAME[int(prop_id)], core)

        return layer

    def weapon(self, w: WEngine) -> PropLayer:
        layer = PropLayer()

        layer.add(PROP_ID_TO_NAME[w.main_stat.type.value], w.main_stat.value)
        layer.add(PROP_ID_TO_NAME[w.sub_stat.type.value], w.sub_stat.value)

        return layer

    def discs(self, ds: list[DriveDisc]) -> PropLayer:
        layer = PropLayer()

        for d in ds:
            layer.add(PROP_ID_TO_NAME[d.main_stat.type.value], d.main_stat.value)

            for sub in d.sub_stats:
                layer.add(PROP_ID_TO_NAME[sub.type.value], sub.value)

        return layer

    def disc_sets(self, ds: list[DriveDisc]) -> PropLayer:
        sets: defaultdict[int, int] = defaultdict(int)
        for d in ds:
            sets[d.set_id] += 1

        layer = PropLayer()
        for set_id, count in sets.items():
            if count < 2:
                continue

            set_data = self._assets.equipments["Suits"][str(set_id)]
            bonus_props: dict[str, int] = set_data["SetBonusProps"]
            for prop_id, value in bonus_props.items():
                layer.add(PROP_ID_TO_NAME[int(prop_id)], value)

        return layer

    def corrections(self, a: Agent, prop_state: PropState) -> PropLayer:
        layer = PropLayer()
        prop_sum = prop_state.sum()

        enh = a.core_skill_level_num
        if a.id == 1121:
            # Ben's initial ATK increases along with his initial DEF. He gains X% of his initial DEF as ATK.
            layer.add(
                "Atk_Delta",
                math.floor(
                    math.floor(prop_sum.defense * (0.4, 0.46, 0.52, 0.6, 0.66, 0.72, 0.8)[enh])
                ),
            )

        if a.id == 1371:
            # Yixuan gains extra Sheer Force based on her Max HP, with every 1 point of Max HP increasing Sheer Force by 0.1.
            layer.add("SkipDefAtk_Delta", math.floor(prop_sum.max_hp * 0.1))

        if a.specialty is ProfessionType.RUPTURE:
            # Rupture characters gain extra Sheer Force based on their ATK, with every 1 point of ATK increasing Sheer Force by 0.3.
            layer.add("SkipDefAtk_Delta", math.floor(math.floor(prop_sum.atk) * 0.3))

        return layer


class PropLayer:  # noqa: PLR0904
    def __init__(self) -> None:
        self.props = DEFAULT_PROPS.copy()

    def add(self, name: str, value: float) -> None:
        if name not in self.props:
            msg = f"Invalid stat name: {name!r}"
            raise ValueError(msg)

        self.props[name] += value

    @property
    def max_hp(self) -> float:
        return (
            self.props["HpMax_Base"]
            + math.ceil(self.props["HpMax_Base"] * (self.props["HpMax_Ratio"] / 10_000))
            + self.props["HpMax_Delta"]
        )

    @property
    def atk(self) -> float:
        return (
            self.props["Atk_Base"] * (1 + self.props["Atk_Ratio"] / 10_000)
            + self.props["Atk_Delta"]
        )

    @property
    def defense(self) -> float:
        return (
            self.props["Def_Base"] * (1 + self.props["Def_Ratio"] / 10_000)
            + self.props["Def_Delta"]
        )

    @property
    def break_stun(self) -> float:
        return self.props["BreakStun_Base"] * (1 + self.props["BreakStun_Ratio"] / 10_000)

    @property
    def crit(self) -> float:
        return self.props["Crit_Base"] + self.props["Crit_Delta"]

    @property
    def crit_dmg(self) -> float:
        return self.props["CritDmg_Base"] + self.props["CritDmg_Delta"]

    @property
    def pen_ratio(self) -> float:
        return self.props["PenRatio_Base"] + self.props["PenRatio_Delta"]

    @property
    def pen_delta(self) -> float:
        return self.props["PenDelta_Base"] + self.props["PenDelta_Delta"]

    @property
    def sp_recover(self) -> float:
        return (
            self.props["SpRecover_Base"] * (1 + self.props["SpRecover_Ratio"] / 10_000)
            + self.props["SpRecover_Delta"]
        )

    @property
    def element_mystery(self) -> float:
        return self.props["ElementMystery_Base"] + self.props["ElementMystery_Delta"]

    @property
    def element_abnormal_power(self) -> float:
        return (
            self.props["ElementAbnormalPower_Base"]
            * (1 + self.props["ElementAbnormalPower_Ratio"] / 10_000)
            + self.props["ElementAbnormalPower_Delta"]
        )

    @property
    def added_damage_ratio_physics(self) -> float:
        return (
            self.props["AddedDamageRatio_Physics_Base"]
            + self.props["AddedDamageRatio_Physics_Delta"]
        )

    @property
    def added_damage_ratio_fire(self) -> float:
        return self.props["AddedDamageRatio_Fire_Base"] + self.props["AddedDamageRatio_Fire_Delta"]

    @property
    def added_damage_ratio_ice(self) -> float:
        return self.props["AddedDamageRatio_Ice_Base"] + self.props["AddedDamageRatio_Ice_Delta"]

    @property
    def added_damage_ratio_elec(self) -> float:
        return self.props["AddedDamageRatio_Elec_Base"] + self.props["AddedDamageRatio_Elec_Delta"]

    @property
    def added_damage_ratio_ether(self) -> float:
        return (
            self.props["AddedDamageRatio_Ether_Base"] + self.props["AddedDamageRatio_Ether_Delta"]
        )

    @property
    def skip_def_atk(self) -> float:
        return self.props["SkipDefAtk_Base"] + self.props["SkipDefAtk_Delta"]

    @property
    def skip_def_damage_ratio(self) -> float:
        return self.props["SkipDefDamageRatio_Base"] + self.props["SkipDefDamageRatio_Delta"]

    @property
    def rp_recover(self) -> float:
        return (
            self.props["RpRecover_Base"] * (1 + self.props["RpRecover_Ratio"] / 10_000)
            + self.props["RpRecover_Delta"]
        )


class PropState:
    def __init__(self) -> None:
        self.layers: list[PropLayer] = []

    def add(self, layer: PropLayer) -> None:
        self.layers.append(layer)

    def sum(self) -> PropLayer:
        sum_layer = PropLayer()
        for layer in self.layers:
            for key, value in layer.props.items():
                sum_layer.add(key, value)

        return sum_layer
